
import math, pathlib

import xml.etree.ElementTree as ET

from base64 import standard_b64encode


def getfloat(value):
    """The INDI spec specifies several different number formats, given a number
       in any of these formats, this returns a float.
       If an error occurs while parsing the number, a TypeError exception is raised."""
    try:
        if isinstance(value, float):
            return value
        if isinstance(value, int):
            return float(value)
        if not isinstance(value, str):
            raise TypeError
        # negative is True, if the value is negative
        value = value.strip()
        negative = value.startswith("-")
        if negative:
            value = value.lstrip("-")
        # Is the number provided in sexagesimal form?
        if value == "":
            parts = ["0", "0", "0"]
        elif " " in value:
            parts = value.split(" ")
        elif ":" in value:
            parts = value.split(":")
        elif ";" in value:
            parts = value.split(";")
        else:
            # not sexagesimal
            parts = [value, "0", "0"]
        if len(parts) > 3:
            raise TypeError
        # Any missing parts should have zero
        if len(parts) == 1:
            parts.append("0")
            parts.append("0")
        if len(parts) == 2:
            parts.append("0")
        assert len(parts) == 3
        # a part could be empty string, ie if 2:5: is given
        numbers = list(float(x) if x else 0.0 for x in parts)
        floatvalue = numbers[0] + (numbers[1]/60) + (numbers[2]/3600)
        if negative:
            floatvalue = -1 * floatvalue
    except Exception:
        raise TypeError("Unable to parse number value")
    return floatvalue



class PropertyMember:
    "Parent class of SwitchMember etc"

    def __init__(self, name, label=None):
        if not name.isascii():
            raise ValueError("Member name should be all ascii characters.")
        self.name = name
        if label:
            self.label = label
        else:
            self.label = name
        # self.changed is a flag to indicate the value has changed
        # initially start with all values have changed
        self.changed = True

    def checkvalue(self, value, allowed):
        "allowed is a list of values, checks if value is in it"
        if value not in allowed:
            raise ValueError(f"Value \"{value}\" is not one of {str(allowed).strip('[]')}")
        return value


class SwitchMember(PropertyMember):
    """A SwitchMember can only have one of 'On' or 'Off' values"""

    def __init__(self, name, label=None, membervalue="Off"):
        super().__init__(name, label)
        # membervalue should be either 'Off' or 'On'
        self._membervalue = membervalue

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not value:
            raise ValueError(f"The SwitchMember {self.name} value cannot be empty")
        newvalue = self.checkvalue(value, ['On', 'Off'])
        if self._membervalue != newvalue:
            # when a value has changed, set the changed flag
            self.changed = True
            self._membervalue = newvalue

    def defswitch(self):
        """Returns a defSwitch"""
        xmldata = ET.Element('defSwitch')
        xmldata.set("name", self.name)
        xmldata.set("label", self.label)
        xmldata.text = self._membervalue
        return xmldata

    def oneswitch(self):
        """Returns xml of a oneSwitch"""
        xmldata = ET.Element('oneSwitch')
        xmldata.set("name", self.name)
        xmldata.text = self._membervalue
        return xmldata


class LightMember(PropertyMember):
    """A LightMember can only have one of 'Idle', 'Ok', 'Busy' or 'Alert' values"""

    def __init__(self, name, label=None, membervalue="Idle"):
        super().__init__(name, label)
        self._membervalue = membervalue

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not value:
            raise ValueError(f"The LightMember {self.name} value cannot be empty")
        newvalue = self.checkvalue(value, ['Idle','Ok','Busy','Alert'])
        if self._membervalue != newvalue:
            # when a value has changed, set the changed flag
            self.changed = True
            self._membervalue = newvalue

    def deflight(self):
        """Returns xml of a defLight"""
        xmldata = ET.Element('defLight')
        xmldata.set("name", self.name)
        xmldata.set("label", self.label)
        xmldata.text = self._membervalue
        return xmldata

    def onelight(self):
        """Returns xml of a oneLight"""
        xmldata = ET.Element('oneLight')
        xmldata.set("name", self.name)
        xmldata.text = self._membervalue
        return xmldata


class TextMember(PropertyMember):
    """Contains a text string"""

    def __init__(self, name, label=None, membervalue=""):
        super().__init__(name, label)
        self._membervalue = membervalue

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if self._membervalue != value:
            # when a value has changed, set the changed flag
            self.changed = True
            self._membervalue = value

    def deftext(self):
        """Returns a defText"""
        xmldata = ET.Element('defText')
        xmldata.set("name", self.name)
        xmldata.set("label", self.label)
        xmldata.text = self.membervalue
        return xmldata

    def onetext(self):
        """Returns xml of a oneText"""
        xmldata = ET.Element('oneText')
        xmldata.set("name", self.name)
        xmldata.text = self.membervalue
        return xmldata


class NumberMember(PropertyMember):
    """Contains a number, the attributes inform the client how the number should be
       displayed.

       format is a C printf style format, for example %7.2f means the client should
       display the number string with seven characters (including the decimal point
       as a character and leading spaces should be inserted if necessary), and with
       two decimal digits after the decimal point.

       min is the minimum value

       max is the maximum, if min is equal to max, the client should ignore these.

       step is incremental step values, set to zero if not used.

       The above numbers, and the member value can be given as strings, which explicitly
       controls how numbers are placed in the xml protocol. If given as integers or floats
       they will be converted and set as strings.
    """

    def __init__(self, name, label=None, format='', min='0', max='0', step='0', membervalue='0'):
        super().__init__(name, label)
        self.format = format
        if isinstance(min, str):
            self.min = min
        else:
            self.min = str(min)
        if isinstance(max, str):
            self.max = max
        else:
            self.max = str(max)
        if isinstance(step, str):
            self.step = step
        else:
            self.step = str(step)
        if isinstance(membervalue, str):
            self._membervalue = membervalue
        else:
            self._membervalue = str(membervalue)


    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not isinstance(value, str):
            value = str(value)
        if self._membervalue != value:
            # when a value has changed, set the changed flag
            self.changed = True
            self._membervalue = value


    def getfloatvalue(self):
        """The INDI spec allows a number of different number formats, this method returns
           this members value as a float.
           If an error occurs while parsing the number, a TypeError exception is raised."""
        return getfloat(self._membervalue)


    def getformattedvalue(self):
        """This method returns this members value as a formatted string.
           If an error occurs while parsing the number, a TypeError exception is raised."""
        floatval = getfloat(self._membervalue)
        return self.format_number(floatval)


    def format_number(self, value):
        """This takes a float, and returns a formatted string
        """
        if (not self.format.startswith("%")) or (not self.format.endswith("m")):
            return self.format % value
        # sexagesimal format
        if value<0:
            negative = True
            value = abs(value)
        else:
            negative = False
        # number list will be degrees, minutes, seconds
        number_list = [0,0,0]
        if isinstance(value, int):
            number_list[0] = value
        else:
            # get integer part and fraction part
            fractdegrees, degrees = math.modf(value)
            number_list[0] = int(degrees)
            mins = 60*fractdegrees
            fractmins, mins = math.modf(mins)
            number_list[1] = int(mins)
            number_list[2] = 60*fractmins

        # so number list is a valid degrees, minutes, seconds
        # degrees
        if negative:
            number = f"-{number_list[0]}:"
        else:
            number = f"{number_list[0]}:"
        # format string is of the form  %<w>.<f>m
        w,f = self.format.split(".")
        w = w.lstrip("%")
        f = f.rstrip("m")
        if (f == "3") or (f == "5"):
            # no seconds, so create minutes value
            minutes = float(number_list[1]) + number_list[2]/60.0
            if f == "5":
                number += f"{minutes:04.1f}"
            else:
                number += f"{minutes:02.0f}"
        else:
            number += f"{number_list[1]:02d}:"
            seconds = float(number_list[2])
            if f == "6":
                number += f"{seconds:02.0f}"
            elif f == "8":
                number += f"{seconds:04.1f}"
            else:
                number += f"{seconds:05.2f}"

        # w is the overall length of the string, prepend with spaces to make the length up to w
        w = int(w)
        lth = len(number)
        if w>lth:
            number = " "*(w-lth) + number
        return number

    def defnumber(self):
        """Returns a defNumber"""
        xmldata = ET.Element('defNumber')
        xmldata.set("name", self.name)
        xmldata.set("label", self.label)
        xmldata.set("format", self.format)
        xmldata.set("min", self.min)
        xmldata.set("max", self.max)
        xmldata.set("step", self.step)
        xmldata.text = self._membervalue
        return xmldata

    def onenumber(self):
        """Returns xml of a oneNumber"""
        xmldata = ET.Element('oneNumber')
        xmldata.set("name", self.name)
        xmldata.text = self._membervalue
        return xmldata


class BLOBMember(PropertyMember):
    """Contains a 'binary large object' such as an image. The membervalue
       should be either None, a bytes object, a file-like object, or a path
       to a file.

       Typically membervalue is left at None when creating a member, and only
       set to a value via vector[membername] = membervalue prior to calling
       the vector send_setVectorMembers method.

       blobsize is the size of the BLOB before any compression, if left at
       zero, the length of the BLOB will be used. The member blobsize can be
       set by calling the vector.set_blobsize(membername, blobsize) method.

       The BLOB format should be a string describing the BLOB, such as .jpeg
    """

    def __init__(self, name, label=None, blobsize=0, blobformat='', membervalue=None):
        super().__init__(name, label)
        if not isinstance(blobsize, int):
            raise ValueError(f"The BLOBMember {self.name} blobsize must be an integer object")
        # membervalue can be a byte string, path, string path or file like object
        self._membervalue = membervalue
        self.blobsize = blobsize
        self.blobformat = blobformat

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not value:
            raise ValueError(f"The BLOBMember {self.name} value cannot be empty")
        self._membervalue = value

    def defblob(self):
        """Returns a defBlob, does not contain a membervalue"""
        xmldata = ET.Element('defBLOB')
        xmldata.set("name", self.name)
        xmldata.set("label", self.label)
        return xmldata


    def oneblob(self, value=None):
        """Returns xml of a oneBLOB
           If value is None, then uses membervalue as the value
           otherwise uses the value given"""
        xmldata = ET.Element('oneBLOB')
        xmldata.set("name", self.name)
        # the value set in the xmldata object should be a bytes object
        if value is None:
            bytescontent = self.getbytes(self._membervalue)
        else:
            bytescontent = self.getbytes(value)

        if self.blobformat:
            xmldata.set("format", self.blobformat)
        elif (value is None) or isinstance(value, bytes):
            # no self.blobformat available, if possible get it from self._membervalue
            if isinstance(self._membervalue, pathlib.Path):
                xmldata.set("format", "".join(self._membervalue.suffixes) )
            elif isinstance(self._membervalue, str):
                xmldata.set("format", "".join(pathlib.Path(self._membervalue).suffixes) )
            else:
                xmldata.set("format", "")
        elif isinstance(value, pathlib.Path):
            xmldata.set("format", "".join(value.suffixes) )
        elif isinstance(value, str):
            xmldata.set("format", "".join(pathlib.Path(value).suffixes) )
        else:
            xmldata.set("format", "")
        if not self.blobsize:
            self.blobsize = len(bytescontent)
        xmldata.set("size", str(self.blobsize))
        xmldata.text = standard_b64encode(bytescontent).decode("utf-8")
        return xmldata


    @staticmethod
    def getbytes(value):
        "Given a blob value, as bytes or file path, return the bytes"
        if not value:
            raise ValueError("The BLOB value is empty")
        try:
            if isinstance(value, bytes):
                bytescontent = value
            elif isinstance(value, pathlib.Path):
                bytescontent = value.read_bytes()
            elif hasattr(value, "seek") and hasattr(value, "read") and callable(value.read):
                # a file-like object
                # set seek(0) so is read from start of file
                value.seek(0)
                bytescontent = value.read()
                value.close()
            else:
                # could be a path to a file
                with open(value, "rb") as fp:
                    bytescontent = fp.read()
        except Exception:
            raise ValueError("Unable to read the given BLOB value")

        if not isinstance(bytescontent, bytes):
            raise ValueError("On being read, the BLOB value does not give bytes")
        if not bytescontent:
            raise ValueError("The BLOB value is empty")

        return bytescontent




import asyncio

import xml.etree.ElementTree as ET

from datetime import datetime, timezone

import logging
logger = logging.getLogger(__name__)

# All xml data received from the remote connection should be contained in one of the following tags
TAGS = (b'message',
        b'newTextVector',
        b'newNumberVector',
        b'newSwitchVector',
        b'newBLOBVector',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector',
        b'getProperties'
       )


# Note these are strings, as they are used for checking xmldata.tag values

DEFTAGS = ( 'defSwitchVector',
            'defLightVector',
            'defTextVector',
            'defNumberVector',
            'defBLOBVector'
          )

NEWTAGS = ('newTextVector',
           'newNumberVector',
           'newSwitchVector',
           'newBLOBVector'
          )


# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)

# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)


def _makestart(element):
    "Given an xml element, returns a string of its start, including < tag attributes >"
    attriblist = ["<", element.tag]
    for key,value in element.attrib.items():
        attriblist.append(f" {key}=\"{value}\"")
    attriblist.append(">")
    return "".join(attriblist)


class RemoteConnection:

    def __init__(self, host, port,
                       blob_enable,
                       debug_enable ):

        self.indihost = host
        self.indiport = port

        # These will be created when a connection is made using
        # await asyncio.open_connection(self.indihost, self.indiport)
        self._writer = None
        self._reader = None

        #  These can be True or False.
        self.blob_enable = blob_enable
        self.debug_enable = debug_enable

        # This is a set of devicenames learnt on this remote connection
        # populated as devices are learnt, and used to check for duplicate names
        self.devicenames = set()

        # An object for communicating is set when this remote is added to the server
        self._commsobj = None

        # and shutdown routine sets this to True to stop coroutines
        self._stop = False


    @property
    def connected(self):
        "property showing connected status, True or False"
        if self._writer is None:
            return False
        else:
            return True

    async def _clear_connection(self):
        "Clears a connection"
        try:
            if self._writer is not None:
                self._writer.close()
                await self._writer.wait_closed()
        except Exception:
            logger.exception("Exception report from RemoteConnection._clear_connection method")
        await self.warning(f"Connection closed on {self.indihost}:{self.indiport}")
        self._writer = None
        self._reader = None

    def __contains__(self, item):
        "So a devicename can easily be checked if it is in this driver"
        return item in self.devicenames


    async def _monitor_connection(self):
        """Flag connection made or failed messages into the server"""
        # create a flag so 'remote connection made' message is only set once
        isconnected = False
        while not self._stop:
            await asyncio.sleep(0.1)
            if self.connected:
                if isconnected:
                    continue
                isconnected = True
                # a new connection has been made, send a getProperties down the link
                await self.send_getProperties()
                # broadcast a message to everyone else connected to the server
                if self._commsobj is not None:
                    timestamp = datetime.now(tz=timezone.utc)
                    timestamp = timestamp.replace(tzinfo = None)
                    tstring = timestamp.isoformat(sep='T')
                    messagedata = ET.Element('message')
                    messagedata.set("timestamp", tstring)
                    messagedata.set("message", f"Remote connection made to {self.indihost}:{self.indiport}")
                    await self._commsobj.run_tx(messagedata)
            else:
                # The connection has failed
                if isconnected:
                    isconnected = False
                    self.devicenames.clear()
                    # broadcast a message to everyone else connected to the server
                    if self._commsobj is not None:
                        # send a message to everyone else connected to the server
                        timestamp = datetime.now(tz=timezone.utc)
                        timestamp = timestamp.replace(tzinfo = None)
                        tstring = timestamp.isoformat(sep='T')
                        messagedata = ET.Element('message')
                        messagedata.set("timestamp", tstring)
                        messagedata.set("message", f"Remote connection to {self.indihost}:{self.indiport} lost")
                        await self._commsobj.run_tx(messagedata)


    def shutdown(self):
        "Shuts down the client, sets the flag self._stop to True"
        self._stop = True
        if self._commsobj is not None:
            self._commsobj.shutdown()

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the client"
        return self._stop


    async def warning(self, message):
        """The given string message will be logged at level WARNING,
           and will be sent to all drivers and connections."""
        try:
            logger.warning(message)
            timestamp = datetime.now(tz=timezone.utc)
            timestamp = timestamp.replace(tzinfo=None)
            xmldata = ET.fromstring(f"<message timestamp=\"{timestamp.isoformat(sep='T')}\" message=\"{message}\" />")
            await self._commsobj.run_tx_everywhere(xmldata)
        except Exception :
            logger.exception("Exception report from RemoteConnection.warning method")


    async def _create_connection(self):
        "Create a connection to an INDI port"
        try:
            while not self._stop:
                t1 = None
                t2 = None
                try:
                    # start by openning a connection
                    await self.warning(f"Attempting to connect to {self.indihost}:{self.indiport}")
                    self._reader, self._writer = await asyncio.open_connection(self.indihost, self.indiport)
                    await self.warning(f"Connected to {self.indihost}:{self.indiport}")
                    await self._run_rx()
                except ConnectionRefusedError:
                    await self.warning(f"Connection refused on {self.indihost}:{self.indiport}")
                except ConnectionError:
                    await self.warning(f"Connection Lost on {self.indihost}:{self.indiport}")
                except OSError:
                    await self.warning(f"Connection Error on {self.indihost}:{self.indiport}")
                except Exception:
                    logger.exception(f"Connection Error on {self.indihost}:{self.indiport}")
                    await self.warning("Connection failed")
                await self._clear_connection()
                if self._stop:
                    break
                else:
                    await self.warning("Connection failed, re-trying...")
                # wait five seconds before re-trying, but keep checking
                # that self._stop has not been set
                count = 0
                while not self._stop:
                    await asyncio.sleep(0.5)
                    count += 1
                    if count >= 10:
                        break
        except Exception:
            logger.exception("Exception report from RemoteConnection._create_connection method")
            raise
        finally:
            await self._clear_connection()
            self.shutdown()


    async def send(self, xmldata):
        """Transmits xmldata, which is an xml.etree.ElementTree object out on the remote connection"""
        if not self.connected:
            return
        if self._stop:
            return
        try:
            if not self.blob_enable:
                if (xmldata.tag == "setBLOBVector") or  (xmldata.tag == "newBLOBVector"):
                    # blobs not enabled
                    return
            # send it out on the port
            binarydata = ET.tostring(xmldata)
            self._writer.write(binarydata)
            await self._writer.drain()
            # data has been transmitted
            if logger.isEnabledFor(logging.DEBUG):
                self._logtx(xmldata)
        except Exception:
            logger.exception(f"Sending error from RemoteConnection.send method for {self.indihost}:{self.indiport}")
            await self._clear_connection()


    async def _readdata(self, xmldata):
        """Called from communications object with  xmldata from other drivers
           . clients and connections. Sends it towards the remote connection"""
        if xmldata.tag == "enableBLOB":
            return
        await self.send(xmldata)


    def _logtx(self, xmldata):
        "log tx data with level debug"
        if not self.debug_enable:
            return
        binarydata = ET.tostring(xmldata)
        logger.debug("TX:: " + binarydata.decode())


    def _logrx(self, rxdata):
        "log rx data"
        if not self.debug_enable:
            return
        binarydata = ET.tostring(rxdata)
        logger.debug("RX:: " + binarydata.decode())


    async def _run_rx(self):
        "accept xml.etree.ElementTree data from the connection"
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput
            while self.connected and (not self._stop):
                xmldata = await self._xmlinput()
                if xmldata is None:
                    return
                if (not self.blob_enable) and (xmldata.tag in ("setBLOBVector", "newBLOBVector") ):
                    # blobs not enabled
                    continue


                devicename = xmldata.get("device")
                vectorname = xmldata.get("name")

                if xmldata.tag in DEFTAGS:
                    if (devicename is None) or (vectorname is None):
                        continue
                    if devicename not in self.devicenames:
                        self.devicenames.add(devicename)
                    if xmldata.tag == "defBLOBVector":
                        # every time a defBLOBVector is received, send an enable BLOB instruction
                        senddata = ET.Element('enableBLOB')
                        senddata.set("device", devicename)
                        senddata.set("name", vectorname)
                        if self.blob_enable:
                            senddata.text = "Also"
                        else:
                            senddata.text = "Never"
                        await self.send(senddata)

                # and pass data to other drivers and connections
                await self._commsobj.run_tx(xmldata)
                # log it, then continue with next block
                if logger.isEnabledFor(logging.DEBUG):
                    self._logrx(xmldata)
        except ConnectionError:
            raise
        except Exception:
            logger.exception("Exception report from RemoteConnection._run_rx")
            raise




    async def _xmlinput(self):
        """get received data, parse it, and return it as xml.etree.ElementTree object
           Returns None if notconnected/stop flags arises"""
        message = b''
        messagetagnumber = None
        while self.connected and (not self._stop):
            await asyncio.sleep(0)
            data = await self._datainput()
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if not self.connected:
                return
            if self._stop:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except ET.ParseError:
                       # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except ET.ParseError:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self):
        """Waits for binary string of data ending in > from the port
           Returns None if notconnected/stop flags arises"""
        binarydata = b""
        while self.connected and (not self._stop):
            await asyncio.sleep(0)
            try:
                data = await self._reader.readuntil(separator=b'>')
            except asyncio.LimitOverrunError:
                data = await self._reader.read(n=32000)
            except asyncio.IncompleteReadError:
                binarydata = b""
                await asyncio.sleep(0.1)
                continue
            if not data:
                await asyncio.sleep(0.01)
                continue
            # data received
            if b">" in data:
                binarydata = binarydata + data
                return binarydata
            # data has content but no > found
            binarydata += data
            # could put a max value here to stop this increasing indefinetly


    async def send_getProperties(self):
        """Sends a getProperties request"""
        if self.connected:
            xmldata = ET.Element('getProperties')
            xmldata.set("version", "1.7")
            await self.send(xmldata)


    async def asyncrun(self):
        "Await this method to run the connectiont."
        self._stop = False
        try:

            async with asyncio.TaskGroup() as tg:
                tg.create_task( self._create_connection() )
                tg.create_task( self._monitor_connection() )

        except Exception:
            logger.exception("Remote connection shutdown")
        finally:
            self._stop = True





import logging

from datetime import datetime, timezone

from base64 import standard_b64decode

from collections import UserDict

from .propertymembers import getfloat

logger = logging.getLogger(__name__)


def _parse_timestamp(timestamp_string):
    """Parse a timestamp string and return either None on failure, or a datetime object
       If the given timestamp_string is None, return the datetime for the current time.
       Everything is UTC"""
    if timestamp_string:
        try:
            if '.' in timestamp_string:
                # remove fractional part, not supported by datetime.fromisoformat
                timestamp_string, remainder = timestamp_string.rsplit('.', maxsplit=1)
                if len(remainder) < 6:
                    remainder = "{:<06}".format(remainder)
                elif len(remainder) > 6:
                    remainder = remainder[:6]
                remainder = int(remainder)
                timestamp = datetime.fromisoformat(timestamp_string)
                timestamp = timestamp.replace(microsecond=remainder, tzinfo=timezone.utc)
            else:
                timestamp = datetime.fromisoformat(timestamp_string)
                timestamp = timestamp.replace(tzinfo=timezone.utc)
        except Exception:
            timestamp = None
    else:
        timestamp = datetime.now(tz=timezone.utc)
    return timestamp


class Event:
    "Parent class for events"
    def __init__(self, devicename, vectorname, vector, root):
        self.devicename = devicename
        self.vectorname = vectorname
        self.vector = vector
        self.root = root


class EventException(Exception):
    "Raised if an error occurs when parsing received data"
    pass


class getProperties(Event):
    "defines an event that a getProperties has been received"
    def __init__(self, devicename, vectorname, vector, root):
        super().__init__(devicename, vectorname, vector, root)
        self.timestamp = datetime.now(tz=timezone.utc)


class newVector(Event, UserDict):
    "Parent to new vectors, adds dictionary and self.timestamp"
    def __init__(self, devicename, vectorname, vector, root):
        Event.__init__(self, devicename, vectorname, vector, root)
        UserDict.__init__(self)
        timestamp_string = root.get("timestamp")
        self.timestamp = _parse_timestamp(timestamp_string)

    def __setitem__(self, membername):
        raise KeyError


class newSwitchVector(newVector):
    """An event indicating a newSwitchVector has been received, this is a mapping
       of membername:value, where each value is either On or Off"""

    def __init__(self, devicename, vectorname, vector, root):
        newVector.__init__(self, devicename, vectorname, vector, root)
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneSwitch":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name set in oneSwitch")
                if membername in self.vector:
                    if not member.text:
                        raise EventException("Missing value in oneSwitch")
                    value = member.text.strip()
                    if value == "On":
                        self.data[membername] = "On"
                    elif value == "Off":
                        self.data[membername] = "Off"
                    else:
                        raise EventException("Received invalid value for newSwitchVector")
                else:
                    raise EventException("Received membername not known for newSwitchVector")
            else:
                raise EventException("Received tag not known for newSwitchVector")
        if not self.data:
            raise EventException("No contents received for newSwitchVector")


class newTextVector(newVector):
    """An event indicating a newTextVector has been received, this is a mapping
       of membername:value, where each value is a text string"""


    def __init__(self, devicename, vectorname, vector, root):
        newVector.__init__(self, devicename, vectorname, vector, root)
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneText":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name set in oneText")
                if membername in self.vector:
                    if member.text is None:
                        self.data[membername] = ""
                    else:
                        self.data[membername] = member.text
                else:
                    raise EventException("Received membername not known for newTextVector")
            else:
                raise EventException("Received tag not known for newTextVector")
        if not self.data:
            raise EventException("No contents received for newTextVector")


class newNumberVector(newVector):
    """An event indicating a newNumberVector has been received, this is a mapping
       of membername:value, where each value is a string number, which may be in
       sexagesimal format, and may have newlines appended or prepended."""

    def __init__(self, devicename, vectorname, vector, root):
        newVector.__init__(self, devicename, vectorname, vector, root)
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneNumber":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name set in oneNumber")
                if membername in self.vector:
                    if not member.text:
                        raise EventException("Missing value in oneNumber")
                    if not member.text.strip():
                        raise EventException("Missing value in oneNumber")
                    self.data[membername] = member.text.strip()
                else:
                    raise EventException("Received membername not known for newNumberVector")
            else:
                raise EventException("Received tag not known for newNumberVector")
        if not self.data:
            raise EventException("No contents received for newNumberVector")

    def getfloatvalue(self, membername):
        "Given a received membername in this event, returns the number as a float"
        if membername not in self:
            raise KeyError(f"Unrecognised member: {membername}")
        membervalue = self.data[membername]
        return getfloat(membervalue)

    def getformattedvalue(self, membername):
        "Given a received membername in this event, returns the number as a formatted string"
        if membername not in self:
            raise KeyError(f"Unrecognised member: {membername}")
        if membername not in self.vector:
            raise KeyError("Received membername not known for newNumberVector")
        membervalue = self.data[membername]
        floatval = getfloat(membervalue)
        member = self.vector.data[membername]
        return member.format_number(floatval)




class newBLOBVector(newVector):
    """An event indicating a newBLOBVector has been received, this is a mapping
       of membername:value, where each value is a bytes string.

       This contains a further attribute 'sizeformat' which is a dictionary
       of membername:(membersize, memberformat) these values are provided by the
       client."""

    def __init__(self, devicename, vectorname, vector, root):
        newVector.__init__(self, devicename, vectorname, vector, root)
        # create a dictionary of member name to value,
        # and dictionary sizeformat
        # with key member name and value being a tuple of size, format
        self.sizeformat = {}
        for member in root:
            if member.tag == "oneBLOB":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name set in oneBLOB")
                if membername in self.vector:
                    try:
                        self.data[membername] = standard_b64decode(member.text.encode('ascii'))
                        membersize = int(member.get("size"))
                    except Exception:
                        raise EventException("Unable to decode BLOB")
                    memberformat = member.get("format")
                    if not memberformat:
                        logger.warning("No format received in oneBLOB member of newBLOBVector")
                        memberformat = ""
                    self.sizeformat[membername] = (membersize, memberformat)
                else:
                    raise EventException("Received tag not known for newBLOBVector")
            else:
                raise EventException("Received tag not known for newBLOBVector")
        if not self.data:
            raise EventException("No contents received for newBLOBVector")


class SnoopEvent:
    "Parent class for snoop events"
    def __init__(self, root):
        self.devicename = root.get("device")
        self.vectorname = None
        self.root = root
        timestamp_string = root.get("timestamp")
        self.timestamp = _parse_timestamp(timestamp_string)


class Message(SnoopEvent):
    """This contains attribute 'message' with the message string sent by the remote driver.
       Attribute devicename could be None if the driver is sending a system wide message."""

    def __init__(self, root):
        super().__init__(root)
        self.message = root.get("message", "")


class delProperty(SnoopEvent):
    """The remote driver is instructing the client to delete either a device or a vector property.
       This contains attribute vectorname, if it is None, then the whole device is to be deleted.
       A 'message' attribute contains any message sent by the remote driver with this instruction."""

    def __init__(self, root):
        super().__init__(root)
        if self.devicename is None:
            raise EventException("No devicename given in snooped delProperty")
        self.vectorname = root.get("name")
        self.message = root.get("message", "")


class defVector(SnoopEvent, UserDict):
    "Parent to def vectors, adds a mapping of membername:value"
    def __init__(self, root):
        SnoopEvent.__init__(self, root)
        UserDict.__init__(self)
        self.vectorname = root.get("name")
        if self.vectorname is None:
            raise EventException("No vectorname given in snooped defVector")
        self.label = root.get("label", self.vectorname)
        self.group = root.get("group", "")
        state = root.get("state")
        if not state:
            raise EventException("No state given in snooped defVector")
        if state not in ('Idle','Ok','Busy','Alert'):
            raise EventException("Invalid state given in snooped defVector")
        self.state = state
        self.message = root.get("message", "")

    def __setitem__(self, membername):
        raise KeyError


class defSwitchVector(defVector):

    """The remote driver has sent this to define a switch vector property, it has further
       attributes perm, rule, timeout, and memberlabels which is a dictionary of
       membername:label."""

    def __init__(self, root):
        defVector.__init__(self, root)
        self.perm = root.get("perm")
        if self.perm is None:
            raise EventException("No perm value given in snooped defSwitchVector")
        if self.perm not in ('ro', 'wo', 'rw'):
            raise EventException("Invalid perm value given in snooped defSwitchVector")
        self.rule = root.get("rule")
        if self.rule is None:
            raise EventException("No rule value given in snooped defSwitchVector")
        if self.rule not in ('OneOfMany', 'AtMostOne', 'AnyOfMany'):
            raise EventException("Invalid rule value given in snooped defSwitchVector")
        self.timeout = root.get("timeout", "0")
        # create object dictionary of member name to value
        # and another dictionary of self.memberlabels with key member name and value being label
        self.memberlabels = {}
        for member in root:
            if member.tag == "defSwitch":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped defSwitchVector")
                label = member.get("label", membername)
                self.memberlabels[membername] = label
                value = member.text.strip()
                if value == "On":
                    self.data[membername] = "On"
                elif value == "Off":
                    self.data[membername] = "Off"
                else:
                    raise EventException("Invalid member value given in snooped defSwitchVector")
            else:
                raise EventException("Invalid tag given in snooped defSwitchVector")
        if not self.data:
            raise EventException("No contents given in snooped defSwitchVector")


class defTextVector(defVector):

    """The remote driver has sent this to define a text vector property, it has further
       attributes perm, timeout, and memberlabels which is a dictionary of
       membername:label."""

    def __init__(self, root):
        defVector.__init__(self, root)
        self.perm = root.get("perm")
        if self.perm is None:
            raise EventException("No perm value given in snooped defTextVector")
        if self.perm not in ('ro', 'wo', 'rw'):
            raise EventException("Invalid perm value given in snooped defTextVector")
        self.timeout = root.get("timeout", "0")
        # create object dictionary of member name to value
        # and another dictionary of self.memberlabels with key member name and value being label
        self.memberlabels = {}
        for member in root:
            if member.tag == "defText":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped defTextVector")
                label = member.get("label", membername)
                self.memberlabels[membername] = label
                self.data[membername] = member.text
            else:
                raise EventException("Invalid tag given in snooped defTextVector")
        if not self.data:
            raise EventException("No contents given in snooped defTextVector")


class defNumberVector(defVector):

    """The remote driver has sent this to define a number vector property, it has further
       attributes perm, timeout, and memberlabels which is a dictionary of
       membername:(label, format, min, max, step)."""

    def __init__(self, root):
        defVector.__init__(self, root)
        self.perm = root.get("perm")
        if self.perm is None:
            raise EventException("No perm value given in snooped defNumberVector")
        if self.perm not in ('ro', 'wo', 'rw'):
            raise EventException("Invalid perm value given in snooped defNumberVector")
        self.timeout = root.get("timeout", "0")
        # create object dictionary of member name to value
        # and another dictionary of self.memberlabels with key member name and
        # value being a tuple of (label, format, min, max, step)
        self.memberlabels = {}
        for member in root:
            if member.tag == "defNumber":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped defNumberVector")
                label = member.get("label", membername)
                memberformat = member.get("format")
                if not memberformat:
                    raise EventException("No format string given in snooped defNumberVector")
                membermin = member.get("min")
                if not membermin:
                    raise EventException("No minimum given in snooped defNumberVector")
                membermax = member.get("max")
                if not membermax:
                    raise EventException("No maximum given in snooped defNumberVector")
                memberstep = member.get("step")
                if not memberstep:
                    raise EventException("No step given in snooped defNumberVector")
                self.memberlabels[membername] = (label, memberformat, membermin, membermax, memberstep)
                self.data[membername] = member.text.strip()
            else:
                raise EventException("Invalid tag given in snooped defNumberVector")
        if not self.data:
            raise EventException("No contents given in snooped defNumberVector")

    def getfloatvalue(self, membername):
        "Given a received membername in this event, returns the number as a float"
        if membername not in self:
            raise KeyError(f"Unrecognised member: {membername}")
        membervalue = self.data[membername]
        return getfloat(membervalue)



class defLightVector(defVector):

    """The remote driver has sent this to define a light vector property, it has further
       attribute memberlabels which is a dictionary of membername:label."""

    def __init__(self, root):
        defVector.__init__(self, root)
        # create object dictionary of member name to value
        # and another dictionary of self.memberlabels with key member name and value being label
        self.memberlabels = {}
        for member in root:
            if member.tag == "defLight":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped defLightVector")
                label = member.get("label", membername)
                self.memberlabels[membername] = label
                value = member.text.strip()
                if value not in ('Idle','Ok','Busy','Alert'):
                    raise EventException("Invalid value given in snooped defLightVector")
                self.data[membername] = value
            else:
                raise EventException("Invalid tag given in snooped defLightVector")
        if not self.data:
            raise EventException("No contents given in snooped defLightVector")


class defBLOBVector(SnoopEvent):

    """The remote driver has sent this to define a BLOB vector property, it has further
       attributes perm, timeout, and memberlabels which is a dictionary of
       membername:label.

       However this class does not have an object mapping of member name to value, since
       values are not given in defBLOBVectors"""

    def __init__(self, root):
        SnoopEvent.__init__(self, root)
        if self.devicename is None:
            raise EventException("No devicename given in snooped defBLOBVector")
        self.vectorname = root.get("name")
        if self.vectorname is None:
            raise EventException("No vectorname given in snooped defBLOBVector")
        self.label = root.get("label", self.vectorname)
        self.group = root.get("group", "")
        state = root.get("state")
        if not state:
            raise EventException("No state given in snooped defBLOBVector")
        if state not in ('Idle','Ok','Busy','Alert'):
            raise EventException("Invalid state given in snooped defBLOBVector")
        self.state = state
        self.message = root.get("message", "")
        self.perm = root.get("perm")
        if self.perm is None:
            raise EventException("No perm value given in snooped defBLOBVector")
        if self.perm not in ('ro', 'wo', 'rw'):
            raise EventException("Invalid perm value given in snooped defBLOBVector")
        self.timeout = root.get("timeout", "0")
        # create a dictionary of self.memberlabels with key member name and value being label
        self.memberlabels = {}
        for member in root:
            if member.tag == "defBLOB":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped defBLOBVector")
                label = member.get("label", membername)
                self.memberlabels[membername] = label
            else:
                raise EventException("Invalid tag given in snooped defBLOBVector")
        if not self.memberlabels:
            raise EventException("No member labels given in snooped defBLOBVector")


class setVector(SnoopEvent, UserDict):
    "Parent to set vectors, adds dictionary"
    def __init__(self, root):
        SnoopEvent.__init__(self, root)
        UserDict.__init__(self)
        if self.devicename is None:
            raise EventException("No device name given in snooped setVector")
        self.vectorname = root.get("name")
        if self.vectorname is None:
            raise EventException("No vector name given in snooped setVector")
        state = root.get("state")
        if state and (state in ('Idle','Ok','Busy','Alert')):
            self.state = state
        else:
            self.state = None
        self.message = root.get("message", "")

    def __setitem__(self, membername):
        raise KeyError


class setSwitchVector(setVector):
    """The remote driver is setting a Switch vector property, this
       has further attribute timeout."""

    def __init__(self, root):
        setVector.__init__(self, root)
        self.timeout = root.get("timeout", "0")
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneSwitch":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped setSwitchVector")
                if not member.text:
                    raise EventException("Missing value in snooped oneSwitch")
                value = member.text.strip()
                if value == "On":
                    self.data[membername] = "On"
                elif value == "Off":
                    self.data[membername] = "Off"
                else:
                    raise EventException("Invalid value given in snooped setSwitchVector")
            else:
                raise EventException("Invalid tag given in snooped setSwitchVector")
        if not self.data:
            raise EventException("No contents given in snooped setSwitchVector")


class setTextVector(setVector):

    """The remote driver is setting a Text vector property, this
       has further attribute timeout."""

    def __init__(self, root):
        setVector.__init__(self, root)
        self.timeout = root.get("timeout", "0")
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneText":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped setTextVector")
                if member.text is None:
                    self.data[membername] = ""
                else:
                    self.data[membername] = member.text
            else:
                raise EventException("Invalid tag given in snooped setTextVector")
        if not self.data:
            raise EventException("No contents given in snooped setTextVector")


class setNumberVector(setVector):

    """The remote driver is setting a Number vector property, this
       has further attribute timeout. The number values of the
       membername:membervalue are string values."""

    def __init__(self, root):
        setVector.__init__(self, root)
        self.timeout = root.get("timeout", "0")
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneNumber":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped setNumberVector")
                if not member.text:
                    raise EventException("Missing value in snooped oneNumber")
                if not member.text.strip():
                    raise EventException("Missing value in snooped oneNumber")
                self.data[membername] = member.text.strip()
            else:
                raise EventException("Invalid tag given in snooped setNumberVector")
        if not self.data:
            raise EventException("No contents given in snooped setNumberVector")

    def getfloatvalue(self, membername):
        "Given a received membername in this event, returns the number as a float"
        if membername not in self:
            raise KeyError(f"Unrecognised member: {membername}")
        membervalue = self.data[membername]
        return getfloat(membervalue)


class setLightVector(setVector):

    """The remote driver is setting a Light vector property."""

    def __init__(self, root):
        setVector.__init__(self, root)
        # create a dictionary of member name to value
        for member in root:
            if member.tag == "oneLight":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped setLightVector")
                if not member.text:
                    raise EventException("Missing value in snooped oneLight")
                value = member.text.strip()
                if value not in ('Idle','Ok','Busy','Alert'):
                    raise EventException("Invalid value in snooped oneLight")
                self.data[membername] = value
            else:
                raise EventException("Invalid tag given in snooped setLightVector")
        if not self.data:
            raise EventException("No contents given in snooped setLightVector")


class setBLOBVector(setVector):

    """The remote driver is setting a BLOB vector property, this
       has further attributes timeout and sizeformat which is a dictionary
       of membername:(size, format)."""

    def __init__(self, root):
        setVector.__init__(self, root)
        self.timeout = root.get("timeout", "0")
        # create a dictionary of member name to value
        # and dictionary sizeformat
        # with key member name and value being a tuple of size, format
        self.sizeformat = {}
        for member in root:
            if member.tag == "oneBLOB":
                membername = member.get("name")
                if not membername:
                    raise EventException("No member name given in snooped setBLOBVector")
                if not member.get("size"):
                    raise EventException("No member size given in snooped setBLOBVector")
                memberformat = member.get("format")
                if not memberformat:
                    raise EventException("No member format given in snooped setBLOBVector")
                try:
                    self.data[membername] = standard_b64decode(member.text.encode('ascii'))
                    membersize = int(member.get("size"))
                except Exception:
                    raise EventException("Unable to decode snooped setBLOBVector")
                self.sizeformat[membername] = (membersize, memberformat)
            else:
                raise EventException("Invalid tag given in snooped setBLOBVector")
        if not self.data:
            raise EventException("No contents given in snooped setBLOBVector")




import sys

from .ipydriver import IPyDriver, Device
from .propertyvectors import SwitchVector, LightVector, TextVector, BLOBVector, NumberVector
from .propertymembers import SwitchMember, LightMember, TextMember, BLOBMember, NumberMember, getfloat
from .events import getProperties, newSwitchVector, newTextVector, newNumberVector, newBLOBVector, Message, delProperty, defSwitchVector, defTextVector, defNumberVector, defLightVector, defBLOBVector, setSwitchVector, setTextVector, setNumberVector, setLightVector, setBLOBVector
from .ipyserver import IPyServer

if sys.version_info < (3, 11):
    raise ImportError('indipydriver requires Python >= 3.11')

version = "2.5.0"


__all__ = ["version", "IPyDriver", "Device", "getfloat", "IPyServer",
           "SwitchVector", "LightVector", "TextVector", "BLOBVector", "NumberVector",
           "SwitchMember", "LightMember", "TextMember", "BLOBMember", "NumberMember",
           "getProperties", "newSwitchVector", "newTextVector", "newNumberVector", "newBLOBVector",
           "Message", "delProperty", "defSwitchVector", "defTextVector", "defNumberVector", "defLightVector", "defBLOBVector",
           "setSwitchVector", "setTextVector", "setNumberVector", "setLightVector", "setBLOBVector"]




import asyncio, copy

import xml.etree.ElementTree as ET

import logging
logger = logging.getLogger(__name__)


# All xml data sent from the driver should be contained in one of the following tags
TAGS = (b'message',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector',
        b'getProperties'       # for snooping
       )

# Note these are strings, as they are used for checking xmldata.tag values

DEFTAGS = ( 'defSwitchVector',
            'defLightVector',
            'defTextVector',
            'defNumberVector',
            'defBLOBVector'
          )


NEWTAGS = ('newTextVector',
           'newNumberVector',
           'newSwitchVector',
           'newBLOBVector'
          )

# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)

# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)


class ExVector:

    def __init__(self, name, deftag):
        "Object which only holds a name and vectortype"
        self.name = name
        self.vectortype = deftag[3:]


class ExDriver:

    def __init__(self, program, *args, debug_enable=False):
        "Executes a third party indi driver, communicates by stdin stdout"

        self.program = program
        self.args = args
        self.debug_enable = debug_enable
        self.proc = None

        # An object for communicating is set when this driver is added to the server
        self._commsobj = None

        # This dictionary will be populated with devicename:{vectorname:vector}
        # where vector is an ExVector object
        self.devicenames = {}

        self.snoopall = False           # gets set to True if it is snooping everything
        self.snoopdevices = set()       # gets set to a set of device names
        self.snoopvectors = set()       # gets set to a set of (devicename,vectorname) tuples

        self._remainder = b""    # Used to store intermediate data
        self._stop = False       # Gets set to True to stop communications


    def shutdown(self):
        self._stop = True
        if self._commsobj is not None:
            self._commsobj.shutdown()
        if self.proc is not None:
            self.proc.terminate()


    def __contains__(self, item):
        "So a devicename can easily be checked if it is in this driver"
        return item in self.devicenames


    async def _readdata(self, root):
        """Called from communications object with received xmldata
           and sends it towards the driver"""

        if self._stop:
            return

        if root.tag == "enableBLOB":
            return

        if root.tag == "getProperties":
            await self.xml_to_ext(root)
            return

        devicename = root.get("device")
        propertyname = root.get("name")

        if root.tag in NEWTAGS:
            if devicename is None:
                # invalid
                return
            if devicename in self.devicenames:
                await self.xml_to_ext(root)
                return
            else:
                # not for this driver
                return

        # so not in NEWTAGS
        # Check if this driver is snooping on this device/vector
        if self.snoopall:
            await self.xml_to_ext(root)
        elif devicename and (devicename in self.snoopdevices):
            await self.xml_to_ext(root)
        elif devicename and propertyname and ((devicename, propertyname) in self.snoopvectors):
            await self.xml_to_ext(root)


    async def xml_to_ext(self, root):
        "Pass data to the external driver"

        binarydata = ET.tostring(root)

        # log the received data
        if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
            logger.debug(f"RX:: {binarydata.decode('utf-8')}")

        binarydata += b"\n"
        self.proc.stdin.write(binarydata)
        await self.proc.stdin.drain()



    async def _run_tx(self):
        "Get data from driver and send it towards the server"
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and append it to self.writerque
            while not self._stop:
                txdata = await self._xmlinput()
                if txdata is None:
                    return
                # get block of xml.etree.ElementTree data
                devicename = txdata.get("device")
                if devicename and txdata.tag in DEFTAGS:
                    # its a definition
                    if devicename not in self.devicenames:
                        self.devicenames[devicename] = {}
                    vectorname = txdata.get("name")
                    if vectorname:
                        if vectorname not in self.devicenames[devicename]:
                            # add this vector to the self.devicenames[devicename] dictionary
                            self.devicenames[devicename][vectorname] = ExVector(vectorname, txdata.tag)
                # check for a getProperties being sent, record what is being snooped
                if txdata.tag == "getProperties":
                    vectorname = txdata.get("name")
                    if devicename is None:
                        self.snoopall = True
                    elif vectorname is None:
                        self.snoopdevices.add(devicename)
                    else:
                        self.snoopvectors.add((devicename,vectorname))

                # transmit this data towards the server
                await self._commsobj.run_tx(txdata)

                if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
                    binarydata = ET.tostring(txdata)
                    logger.debug(f"TX:: {binarydata.decode('utf-8')}")

        except Exception:
            logger.exception("Exception report from ExDriver._run_tx")
            raise


    async def _xmlinput(self):
        """get data from driver, parse it, and return it as xml.etree.ElementTree object
           Returns None if stop flags arises"""
        message = b''
        messagetagnumber = None
        while not self._stop:
            data = await self._datainput()
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if self._stop:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except Exception:
                        # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except Exception:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self):
        """Waits for binary string of data ending in > from the driver
           Returns None if stop flags arises"""
        remainder = self._remainder
        if b">" in remainder:
            # This returns with binary data ending in > as long
            # as there are > characters in self._remainder
            binarydata, self._remainder = remainder.split(b'>', maxsplit=1)
            binarydata += b">"
            return binarydata
        # As soon as there are no > characters left in self._remainder
        # get more data from the driver
        while not self._stop:
            await asyncio.sleep(0)
            indata = await self.proc.stdout.read(100)
            if not indata:
                await asyncio.sleep(0.02)
                continue
            remainder += indata
            if b">" in indata:
                binarydata, self._remainder = remainder.split(b'>', maxsplit=1)
                binarydata += b">"
                return binarydata


    async def _run_err(self):
        """gets binary string of data from exdriver proc.stderr
           and logs it to logging.error."""

        while not self._stop:
            bindata = await self.proc.stderr.readline()
            if not bindata:
                await asyncio.sleep(0.02)
                continue
            logger.error(bindata.decode('utf-8'))


    async def asyncrun(self):
        "Runs the external driver"

        self._stop = False

        try:

            self.proc = await asyncio.create_subprocess_exec(self.program,
                                                             *self.args,
                                                       stdin=asyncio.subprocess.PIPE,
                                                       stdout=asyncio.subprocess.PIPE,
                                                       stderr=asyncio.subprocess.PIPE)

            # wait for external program to start
            await asyncio.sleep(0.1)
            # send a getProperties into the driver
            await self._readdata( ET.fromstring("""<getProperties version="1.7" />""") )

            async with asyncio.TaskGroup() as tg:
                tg.create_task( self._run_tx() )           # Get data from driver and send it towards the server
                tg.create_task( self._run_err() )          # data from exdriver proc.stderr and logs it to logging.error
        except Exception:
            logger.exception("Driver shutdown")
        finally:
            self._stop = True





import asyncio, logging, time

from datetime import datetime, timezone

import xml.etree.ElementTree as ET

from .ipydriver import IPyDriver

from .remote import RemoteConnection

from .exdriver import ExDriver

logger = logging.getLogger(__name__)


# All xml data should be contained in one of the following tags
TAGS = (b'getProperties',
        b'newTextVector',
        b'newNumberVector',
        b'newSwitchVector',
        b'newBLOBVector',
        b'enableBLOB',
        b'message',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector'
       )


# Note these are strings, as they are used for checking xmldata.tag values

DEFTAGS = ( 'defSwitchVector',
            'defLightVector',
            'defTextVector',
            'defNumberVector',
            'defBLOBVector'
          )

NEWTAGS = ('newTextVector',
           'newNumberVector',
           'newSwitchVector',
           'newBLOBVector'
          )


# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)


# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)



def cleanque(que):
    "Clears out a que"
    try:
        while True:
            que.get_nowait()
            que.task_done()
    except asyncio.QueueEmpty:
        # que is now empty, nothing further to do
        pass


class IPyServer:

    """Once an instance of this class is created, the asyncrun method
       should be awaited which will open a port, and the INDI service
       will be available for clients to connect.

       drivers are IPyDriver objects this driver handles,
       host and port are "localhost" and 7624 as default

       maxconnections is the number of simultaneous client connections
       accepted, with a default of 5. The number given should be
       between 1 and 10 inclusive.

       If, prior to asyncrun being awaited, the add_remote method is called,
       then a connection will be made to a remote INDI server and any of its
       drivers.

       The add_remote method can be called multiple times to create
       connections to different servers making a branching
       tree of servers and drivers.

       The add_exdriver method can be called to run an executable driver and
       this server will communicate to it by stdin, stdout and stderr, therefore
       ipyserver can act as a general INDI server for third party drivers as
       well as ipydriver instances.
       """


    def __init__(self, *drivers, host="localhost", port=7624, maxconnections=5):

        self.drivers = list(drivers)
        self.host = host
        self.port = port

        self.keepalive = 30
        # If set to an integer number of seconds, should no traffic pass through the server
        # for that many seconds, a 'keepalive' message will be broadcast

        # all data is sent on one que with format (con_id, xmldata)
        # con_id is used to stop a transmitter receiving its own transmitted data

        self.xml_data_que = asyncio.Queue(50)

        self.con_id = 0

        if maxconnections<1 or maxconnections>10:
            raise ValueError("maxconnections should be a number between 1 and 10 inclusive.")
        self.maxconnections = maxconnections

        # this is a dictionary of device name to device
        self.devices = {}

        # self.remotes is a list of RemoteConnection objects running connections to remote servers
        # this list is populated by calling self.add_remote(host, port, debug_enable)
        self.remotes = []

        # self.exdrivers is a list of ExDriver objects running external drivers
        # this list is populated by calling self.add_exdriver(program, *args, debug_enable=False)
        self.exdrivers = []

        for driver in self.drivers:
            if not isinstance(driver, IPyDriver):
                raise TypeError("The drivers set in IPyServer must all be IPyDrivers")
            for devicename in driver:
                if devicename in self.devices:
                    # duplicate devicename
                    raise ValueError(f"Device name {devicename} is duplicated in the attached drivers.")
            self.devices.update(driver.data)


        self.connectionpool = []
        for c in range(0, maxconnections):
            self.con_id += 1
            self.connectionpool.append(_ClientConnection(self.con_id, self.xml_data_que))

        for driver in self.drivers:
            # an instance of _DriverComms is created for each driver
            self.con_id += 1
            driver._commsobj = _DriverComms(driver, self.con_id, self.xml_data_que)

        # shutdown routine sets this to True to stop coroutines
        self._stop = False
        # this is set when asyncrun is finished
        self.stopped = asyncio.Event()
        self.server = None

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the server"
        return self._stop

    def shutdown(self, shutdownmessage=""):
        """Shuts down the server, sets the flag self._stop to True
           and sends shutdownmessage to logger.error if given"""
        if shutdownmessage:
            logger.error(shutdownmessage)
        self._stop = True
        for driver in self.drivers:
            driver.shutdown()
        for remcon in self.remotes:
            remcon.shutdown()
        for exd in self.exdrivers:
            exd.shutdown()
        for clientconnection in self.connectionpool:
            clientconnection.shutdown()
        if self.server is not None:
            self.server.close()
        self.stopped.set()


    def add_remote(self, host, port, blob_enable=False, debug_enable=False):
        """Adds a connection to a remote server.
           blob_enable can be True or False.
           If True BLOBs and other vectors can all be sent.
           If False, then BLOB traffic will not pass over this link.

           If debug_enable is True, then DEBUG level logging will record xml
           traffic, if False, the xml traffic will not be logged. This can be
           used to prevent multiple such connections all logging xml traffic together."""


        remcon = RemoteConnection( host=host, port=port,
                                   blob_enable = blob_enable,
                                   debug_enable = debug_enable )

        # Create a DriverComms object
        self.con_id += 1
        remcon._commsobj = _DriverComms(remcon, self.con_id, self.xml_data_que)
        # store this object
        self.remotes.append(remcon)


    def add_exdriver(self, program, *args, debug_enable=False):
        """Adds an executable driver program, runs it and communicates to it via stdin, stdout
           and stderr. Then serves the driver, and any others added, by the listening port.
           args is used for the program arguments if any.
           Any program output on stderr will be logged at level ERROR.

           If debug_enable is True, then DEBUG level logging will record xml
           traffic, if False, the xml traffic will not be logged. This can be
           used to prevent multiple drivers all logging xml traffic together."""
        exd = ExDriver(program, *args, debug_enable=debug_enable)
        # Create a DriverComms object
        self.con_id += 1
        exd._commsobj = _DriverComms(exd, self.con_id, self.xml_data_que)
        # store this object
        self.exdrivers.append(exd)



    async def _runserver(self):
        "Runs the server on the given host and port"
        logger.info(f"{self.__class__.__name__} listening on {self.host} : {self.port}")
        self.server = await asyncio.start_server(self.handle_data, self.host, self.port)
        try:
            async with self.server:
                await self.server.serve_forever()
        finally:
            self.shutdown()


    async def handle_data(self, reader, writer):
        "Used by asyncio.start_server, called to handle a client connection"

        try:
            for clientconnection in self.connectionpool:
                if not clientconnection.connected:
                    # this clientconnection is available
                    await clientconnection.handle_data(reader, writer)
                    break
        except Exception:
            # Call has dropped
            logger.info("Call dropped")
        try:
            # call dropped or no clientconnection is available
            writer.close()
            await writer.wait_closed()
        except BrokenPipeError:
            # avoid broken pipe error being displayed
            pass


    async def asyncrun(self):
        """await this to operate the server together with its
           drivers and any remote connections."""

        self._stop = False

        try:
            async with asyncio.TaskGroup() as tg:
                for driver in self.drivers:
                    tg.create_task( driver.asyncrun() )
                for exdriver in self.exdrivers:
                    tg.create_task( exdriver.asyncrun() )
                for remcon in self.remotes:
                    tg.create_task( remcon.asyncrun() )
                tg.create_task( self._runserver() )
                tg.create_task( self._broadcast() )
        except Exception:
            logger.exception("Exception in IPyServer.asyncrun")
        finally:
            self.shutdown()


    async def _broadcast(self):
        "Get items from the que, and broadcast to drivers"
        timeout = time.time()+self.keepalive
        try:
            while not self._stop:
                if self.keepalive:
                    current_time = time.time()
                    if current_time > timeout:
                        timeout = current_time + self.keepalive
                        await self._send_keepalive()
                try:
                    quedata = await asyncio.wait_for(self.xml_data_que.get(), 0.5)
                except asyncio.TimeoutError:
                    continue
                self.xml_data_que.task_done()
                if self._stop:
                    return
                con_id, xmldata = quedata
                # data to send, therefore increase keepalive timeout
                # however, do not increase it if this is a setBLOBVector, as these may be blocked
                if xmldata.tag != "setBLOBVector":
                    timeout = time.time()+self.keepalive
                async with asyncio.TaskGroup() as tg:
                    for driver in self.drivers:
                        # send data to the drivers
                        tg.create_task( driver._commsobj.driver_rx(con_id, xmldata) )
                    for exdriver in self.exdrivers:
                        # send data to the external drivers
                        tg.create_task( exdriver._commsobj.driver_rx(con_id, xmldata) )
                    for remcon in self.remotes:
                        tg.create_task( remcon._commsobj.driver_rx(con_id, xmldata) )
                    for clientconnection in self.connectionpool:
                        # send data out to clients
                        tg.create_task( clientconnection._client_tx(con_id, xmldata) )
        finally:
            self.shutdown()


    async def _send_keepalive(self):
        "This transmits a keep-alive to the client connections"
        timestamp = datetime.now(tz=timezone.utc).replace(tzinfo = None)
        xmldata = ET.Element('message')
        xmldata.set("timestamp", timestamp.isoformat(sep='T'))
        xmldata.set("message", "Keep alive message")
        async with asyncio.TaskGroup() as tg:
            for remcon in self.remotes:
                # send data out to remote connections
                tg.create_task( remcon._commsobj.driver_rx(0, xmldata) )
            for clientconnection in self.connectionpool:
                # send data out to clients
                tg.create_task( clientconnection._client_tx(0, xmldata) )



    async def send_message(self, message, timestamp=None):
        """Send system wide message, timestamp should normally not be set, if
           given, it should be a datetime.datetime object with tz set to timezone.utc"""
        if self._stop:
            return
        if not timestamp:
            timestamp = datetime.now(tz=timezone.utc).replace(tzinfo = None)
        else:
            if not isinstance(timestamp, datetime):
                # invalid timestamp given
                return
            if timestamp.tzinfo is not None:
                if timestamp.tzinfo == timezone.utc:
                    timestamp = timestamp.replace(tzinfo = None)
                else:
                    # invalid timestamp
                    return
        xmldata = ET.Element('message')
        xmldata.set("timestamp", timestamp.isoformat(sep='T'))
        xmldata.set("message", message)
        # send with a con_id of zero, so it is sent everywhere
        await self.xml_data_que.put( (0, xmldata) )


class _DriverComms:

    """An instance of this is created for each driver.  Any data the driver
       wishes to send will be transmitted via run_tx
       Any data received on self.xml_data_que will be sent to the
       driver by calling its driver.receivedata method"""

    def __init__(self, driver, con_id, xml_data_que):

        # This object is attached to this driver
        self.driver = driver
        self.con_id = con_id
        self.xml_data_que = xml_data_que


    def shutdown(self):
        "Called by driver on shutdown, used by _STDINOUT but not relevant here"
        pass


    async def run_rx(self):
        "Called by driver on running, used by _STDINOUT but not relevant here"
        pass


    async def driver_rx(self, con_id, xmldata):
        "Gets data from xml_data_que, and sends it to the driver"
        if self.con_id == con_id:
            # do not rx data this driver is transmitting
            return

        if xmldata.tag == "getProperties":
            version = xmldata.get("version")
            if version != "1.7":
                return

        # check for incoming duplicates
        if xmldata.tag in DEFTAGS:
            devicename = xmldata.get("device")
            if devicename is None:
                # invalid definition
                return
            if devicename in self.driver:
                # duplicate address
                logger.error(f"Duplicate address: Received a definition of device {devicename}")
                self.driver.shutdown()
                raise RuntimeError(f"Device name {devicename} is duplicated in the attached drivers.")


        # call the drivers receive data function
        await self.driver._readdata(xmldata)


    async def run_tx(self, xmldata):
        """Called by the driver, places data generated by the driver into xml_data_que with this connection id"""
        await self.xml_data_que.put( (self.con_id, xmldata) )


    async def run_tx_everywhere(self, xmldata):
        """Called to send data down every connection,places data generated by the driver into xml_data_que with zero connection id"""
        await self.xml_data_que.put( (0, xmldata) )




class _ClientConnection:

    "Handles a client connection"

    def __init__(self, con_id, xml_data_que):

        # number identifying this connection
        self.con_id = con_id

        self.xml_data_que = xml_data_que

        # self.connected is True if this pool object is running a connection
        self.connected = False
        self._remainder = b""    # Used to store intermediate data
        self.sendchecker = None

        self.rx = None


    def shutdown(self):
        "Shuts down the connection"
        self.connected = False

    async def handle_data(self, reader, writer):
        "Used by asyncio.start_server, called to handle a client connection"
        self.connected = True
        self._remainder = b""    # Used to store intermediate data
        self.writer = writer
        self.reader = reader
        self.sendchecker = SendChecker()
        addr = writer.get_extra_info('peername')
        logger.info(f"Connection received from {addr} on connection {self.con_id}")
        try:
            await self._client_rx()
        except ConnectionError:
            pass
        finally:
            self.shutdown()
        logger.info(f"Connection from {addr} closed")


    async def _client_tx(self, con_id, xmldata):
        "Sends data from port to client"
        if self.con_id == con_id:
            # do not tx data this driver is receiving
            return
        if not self.connected:
            return
        if not self.sendchecker.allowed(xmldata):
            # this data should not be transmitted, discard it
            return
        try:
            # this data can be transmitted
            binarydata = ET.tostring(xmldata)
            # Send to the port
            self.writer.write(binarydata)
            await self.writer.drain()
        except ConnectionError:
            self.shutdown()



    async def _client_rx(self):
        "Receives data coming in to port from client"

        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and send it to xml_data_que together with connection_id
            while self.connected:
                xmldata = await self._xmlinput()
                if xmldata is None:
                    return
                if xmldata.tag == "enableBLOB":
                    # set permission flags in the sendchecker object
                    self.sendchecker.setpermissions(xmldata)
                    # do not broadcast this, so continue
                    continue
                # pass xmldata to xml_data_que
                await self.xml_data_que.put( (self.con_id, xmldata) )
        except ConnectionError:
            # re-raise this without creating a report, as it probably indicates
            # a normal connection drop
            raise
        except Exception:
            # possibly some other error, so report it
            logger.exception("Exception report from _ClientConnection._client_rx")
            raise


    async def _xmlinput(self):
        """get data from  _datainput, parse it, and return it as xml.etree.ElementTree object
           Returns None if stop flags arises"""
        message = b''
        messagetagnumber = None
        while self.connected:
            await asyncio.sleep(0)
            data = await self._datainput()
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if not self.connected:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except Exception:
                        # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except Exception:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self):
        """Waits for binary string of data ending in > from the port
           Returns None if stop flags arises"""
        binarydata = b""
        while self.connected:
            await asyncio.sleep(0)
            try:
                data = await self.reader.readuntil(separator=b'>')
            except asyncio.LimitOverrunError:
                data = await self.reader.read(n=32000)
            except asyncio.IncompleteReadError:
                binarydata = b""
                await asyncio.sleep(0.1)
                continue
            if not data:
                await asyncio.sleep(0.1)
                continue
            # data received
            if b">" in data:
                binarydata = binarydata + data
                return binarydata
            # data has content but no > found
            binarydata += data
            # could put a max value here to stop this increasing indefinetly



# Command to control whether setBLOBs should be sent to this channel from a given Device. They can
# be turned off completely by setting Never (the default), allowed to be intermixed with other INDI
# commands by setting Also or made the only command by setting Only.

# <!ELEMENT enableBLOB %BLOBenable >
# <!ATTLIST enableBLOB
# device %nameValue; #REQUIRED  name of Device
# name %nameValue; #IMPLIED name of BLOB Property, or all if absent
# >



class SendChecker:
    """Carries the enableBLOB status on a connection, and does checks
       to ensure valid data is being transmitted"""

    def __init__(self):
        "For every device create a dictionary"
        self.devicestatus = {}
        # create a dictionary of devicenames : to devicedict
        # where the devicedict will be {"Default":"Never", "Properties":{}}
        # The Properties value is a dictionary of propertyname:status


    def allowed(self, xmldata):
        "Return True if this xmldata can be transmitted, False otherwise"

        if xmldata.tag in ("getProperties", "delProperty"):
            return True

        if xmldata.tag not in ("defBLOBVector", "setBLOBVector", 'newBLOBVector'):
            # so anything other than a BLOB
            if self.rxonly():
                # Only blobs allowed
                return False
            return True

        # so following checks only apply to BLOB vectors

        devicename = xmldata.get("device")

        if devicename not in self.devicestatus:
            # devicename not recognised, add it
            self.devicestatus[devicename] = {"Default":"Never", "Properties":{}}

        if xmldata.tag == "defBLOBVector":
            return True

        if (xmldata.tag == "setBLOBVector") and (not xmldata):   # empty setBLOBVectors are allowed regardless
            return True

        devicedict = self.devicestatus[devicename]

        # so we have a devicename, get propertyname
        name = xmldata.get("name")

        # so we have a devicename, property name,
        if name and (name in devicedict["Properties"]):
            if devicedict["Properties"][name] == "Never":
                return False
            else:
                return True
        elif devicedict["Default"] == "Never":
            return False
        else:
            return True

    def setpermissions(self, rxdata):
        "Read the received enableBLOB xml and set permission in self.devicestatus"
        devicename = rxdata.get("device")
        if devicename is None:
            # invalid
            return
        if devicename not in self.devicestatus:
            # devicename not recognised, add it
            self.devicestatus[devicename] = {"Default":"Never", "Properties":{}}

        # get the status of Never, Also, Only
        status = rxdata.text.strip()
        if status not in ("Never", "Also", "Only"):
            # invalid
            return

        devicedict = self.devicestatus[devicename]

        # property name
        name = rxdata.get("name")
        if name is None:
            # This applies to the device rather than to a particular property
            devicedict["Default"] = status
        else:
            # add it to devicedict, and hence to self.devicestatus
            devicedict["Properties"][name] = status

    def rxonly(self):
        "Returns True if any device or property has been set to BLOBs only"
        for devicedict in self.devicestatus.values():
            if devicedict["Default"] == "Only":
                return True
            properties = devicedict["Properties"]
            for status in properties.values():
                if status == "Only":
                    return True
        return False





import collections, asyncio, sys, time, os, fcntl, logging

import xml.etree.ElementTree as ET

from . import events
from .propertyvectors import timestamp_string
from .propertymembers import getfloat


logger = logging.getLogger(__name__)


# All xml data received from the client, or from snooped devices should be contained in one of the following tags
TAGS = (b'getProperties',
        b'newTextVector',
        b'newNumberVector',
        b'newSwitchVector',
        b'newBLOBVector',
        b'enableBLOB',
        # below are tags from snooped devices
        b'message',
        b'delProperty',
        b'defSwitchVector',
        b'setSwitchVector',
        b'defLightVector',
        b'setLightVector',
        b'defTextVector',
        b'setTextVector',
        b'defNumberVector',
        b'setNumberVector',
        b'defBLOBVector',
        b'setBLOBVector'
       )

# Note these are strings, as they are used for checking xmldata.tag values

DEFTAGS = ( 'defSwitchVector',
            'defLightVector',
            'defTextVector',
            'defNumberVector',
            'defBLOBVector'
          )


# _STARTTAGS is a tuple of ( b'<defTextVector', ...  ) data received will be tested to start with such a starttag
_STARTTAGS = tuple(b'<' + tag for tag in TAGS)


# _ENDTAGS is a tuple of ( b'</defTextVector>', ...  ) data received will be tested to end with such an endtag
_ENDTAGS = tuple(b'</' + tag + b'>' for tag in TAGS)


def _makestart(element):
    "Given an xml element, returns a string of its start, including < tag attributes >"
    attriblist = ["<", element.tag]
    for key,value in element.attrib.items():
        attriblist.append(f" {key}=\"{value}\"")
    attriblist.append(">")
    return "".join(attriblist)


class IPyDriver(collections.UserDict):

    """A subclass of this should be created with methods written
       to control your device.

       devices are Device objects this driver handles.

       You may optionally include named arguments of any instrumentation
       objects that may be useful to you, these will be available in
       the dictionary self.driverdata.

       This object is also a mapping, of devicename:deviceobject
       """

    @staticmethod
    def indi_number_to_float(value):
        """The INDI spec allows a number of different number formats, given any number string, this returns a float.
           If an error occurs while parsing the number, a TypeError exception is raised."""
        return getfloat(value)


    def __init__(self, *devices, **driverdata):
        super().__init__()

        # self.data is defined as empty dictionary in UserDict
        for device in devices:
            devicename = device.devicename
            if devicename in self.data:
                # duplicate devicename
                raise ValueError(f"Device name {devicename} is duplicated in this driver.")
            self.data[devicename] = device
            # set driver into devices
            device.driver = self
            # set driver into vectors
            for pv in device.values():
                pv.driver = self

        # self.data is used by UserDict, it will become
        # a dictionary of {devicename:device, ... }

        # dictionary of optional data
        self.driverdata = driverdata


        # initial method of communications, if None it will default to
        # stdin and stdout
        self._commsobj = None

        # These set the remote traffic which this driver is snooping
        # initially the driver is not snooping anything, until it sends
        # a getProperties
        self.snoopall = False           # gets set to True if it is snooping everything
        self.snoopdevices = set()       # gets set to a set of device names

        self.snoopvectors = {}
        # The keys of self.snoopvectors will be tuples (devicename,vectorname)
        # of vectors that are to be snooped
        # The values will be either None or lists of [timeout, timestamp]

        # timeout is integer seconds set by the snoop() method
        # timestamp is updated whenever snoop data from devicename,vectorname
        # is received.
        # The coroutine _monitorsnoop Checks if current time is greater than
        # timeout+timestamp, and if it is, sends a getproperties

        self.debug_enable = False
        # If True, xmldata will be logged at DEBUG level

        self.auto_send_def = True
        # If True, whenever a getProperties event is received, a
        # vector send_defVector() will be called, automatically replying with
        # the vector definition.
        # If set to False, the driver developer will need to implement a send_defVector()
        # in the rxevent method

        self._stop = False
        # shutdown routine sets this to True to stop coroutines

        self.stopped = asyncio.Event()
        # this is set when asyncrun is finished

        # This holds any background tasks which are to be run
        self._background_tasks = set()


    def run_background_task(self, task):
        """This method is typically called from the rxevent method with
           a task to be run in the background, it retains a strong reference
           to the task until it is done."""
        self._background_tasks.add(task)
        # To prevent keeping references to finished tasks forever,
        # make each task remove its own reference from the set after
        # completion:
        task.add_done_callback(self._background_tasks.discard)
        # This task is now running in the background


    def devices(self):
        "Returns a list of device objects"
        return list(self.data.values())


    def shutdown(self):
        "Shuts down the driver, sets the flag self.stop to True"
        self._stop = True
        if self._commsobj is not None:
            self._commsobj.shutdown()
        for device in self.data.values():
            device.shutdown()

    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop



    async def send(self, xmldata):
        "Transmits xmldata, this is an internal method, not normally called by a user."
        if self._stop:
            return

        await self._commsobj.run_tx(xmldata)

        if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
            binarydata = ET.tostring(xmldata)
            logger.debug(f"TX:: {binarydata.decode('utf-8')}")


    def __setitem__(self, devicename):
        raise KeyError


    async def _call_snoopevent(self, event):
        "Update timestamp when snoop data received and call self.snoopevent"
        if event.devicename and event.vectorname:
            # update timestamp in self.snoopvectors
            timedata = self.snoopvectors.get((event.devicename, event.vectorname))
            if timedata is not None:
                timedata[1] = time.time()
        await self.snoopevent(event)


    async def send_message(self, message="", timestamp=None):
        "Send system wide message - without device name"
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_message must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('message')
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.send(xmldata)


    def snoop(self, devicename, vectorname, timeout=30):
        """Call this to snoop on a given devicename, vectorname.
           This will cause a getProperties to be sent, and will also
           send further getProperties every timeout seconds if no snooping
           data is being received from the specified vector.
           This avoids a possible problem where intermediate servers may
           be temporarily turned off, and will lose their instruction to
           broadcast snooping traffic. This method is only applicable when
           snooping on a specific device vector.
           timeout must be an integer equal or greater than 5 seconds."""
        if devicename in self.data:
            logger.error("Cannot snoop on a device already controlled by this driver")
            return
        timeout = int(timeout)
        if timeout < 5:
            logger.error("Snoop timout should be equal or greater than 5 seconds")
            return

        current = time.time()

        # set self.snoopvectors[(devicename,vectorname)] to [timeout, timestamp]

        self.snoopvectors[(devicename,vectorname)] = [timeout, current - timeout + 1]

        # setting timestamp to current - timeout + 1 means that after a second
        # the coroutine _monitorsnoop will think that its own time measurement
        # is greater than the timestamp plus timeout and will send a send_getProperties


    async def _monitorsnoop(self):
        "Checks if any snooping vectors have timed out, if it has, sends getproperties"
        while not self._stop:
            await asyncio.sleep(1)
            if not self.snoopvectors:
                continue
            current = time.time()
            for key, value in self.snoopvectors.items():
                if value is None:
                    continue
                timeout, timestamp = value
                if current > timestamp + timeout:
                    # the timeout has expired, update timestamp and send getproperties
                    value[1] = current
                    await self.send_getProperties(*key)


    async def send_getProperties(self, devicename=None, vectorname=None):
        """Sends a getProperties request - which is used to snoop data from other devices
           on the network, if devicename given, it must not be a device of this driver as
           the point of this is to snoop on remote devices."""
        xmldata = ET.Element('getProperties')
        xmldata.set("version", "1.7")
        if devicename is None:
            await self.send(xmldata)
            self.snoopall = True
            return
        if devicename in self.data:
            logger.error("Cannot snoop on a device already controlled by this driver")
            return
        xmldata.set("device", devicename)
        if vectorname is None:
            await self.send(xmldata)
            self.snoopdevices.add(devicename)
            return
        xmldata.set("name", vectorname)
        await self.send(xmldata)
        # adds tuple (devicename,vectorname) to self.snoopvectors
        if (devicename,vectorname) not in self.snoopvectors:
            self.snoopvectors[(devicename,vectorname)] = None


    async def hardware(self):
        """Override this to operate device hardware, and transmit updates

        For example: call your own code to operate hardware
        then update the appropriate vectors, and send updated
        values to the client using
        await vector.send_setVector()"""
        pass


    async def rxevent(self, event):
        """Override this. On receiving data, this is called, and should
           handle any necessary actions.
           event is an object describing the event, with attributes
           devicename, vectorname, vector, root
           where vector is the properties vector the event refers to, and
           root is an xml.etree.ElementTree object of the received xml"""
        pass


    async def snoopevent(self, event):
        """Override this if this driver is snooping on other devices.
           On receiving snoop data, this is called, and should handle
           any necessary actions.
           event is an object with attributes according to the data received."""
        pass


    async def asyncrun(self):
        """await this to operate the driver, which will then communicate by
           stdin and stdout.

           Do not await this if the driver is being set into IPyServer, in
           that situation the IPyServer will control communications."""

        logger.info(f"Driver {self.__class__.__name__} started")
        self._stop = False

        try:

            if self._commsobj is None:
                self._commsobj = _STDINOUT(self)

            async with asyncio.TaskGroup() as tg:
                tg.create_task( self._commsobj.run_rx() )        # run STDIN communications
                tg.create_task( self.hardware() )                # task to operate device hardware, and transmit updates
                tg.create_task( self._monitorsnoop() )          # task to monitor if a getproperties needs to be sent
        finally:
            self.shutdown()
            self.stopped.set()


    async def _readdata(self, root):
        "Called from communications object with received xmldata"

        client_tags = ("newSwitchVector", "newNumberVector", "newTextVector", "newBLOBVector")
        snoop_tags = ("message", 'delProperty', 'defSwitchVector', 'setSwitchVector', 'defLightVector',
                      'setLightVector', 'defTextVector', 'setTextVector', 'defNumberVector', 'setNumberVector',
                      'defBLOBVector', 'setBLOBVector')
        if self._stop:
            return

        # log the received data
        if logger.isEnabledFor(logging.DEBUG) and self.debug_enable:
            binarydata = ET.tostring(root)
            logger.debug(f"RX:: {binarydata.decode('utf-8')}")

        if root.tag == "getProperties":
            version = root.get("version")
            if version != "1.7":
                return
            # getProperties received with correct version
            devicename = root.get("device")
            # if devicename is None (for all devices), send definition for every property
            if devicename is None:
                try:
                    async with asyncio.TaskGroup() as tg:
                        # all devices respond to this getProperties
                        for dname, device in self.data.items():
                            if not device.enable:
                                continue
                            for pname, pvector in device.items():
                                if not pvector.enable:
                                    continue
                                if self.auto_send_def:
                                    tg.create_task( pvector.send_defVector() )
                                else:
                                    # create event
                                    e = events.getProperties(dname, pname, pvector, root)
                                    tg.create_task( self.rxevent(e) )
                except Exception:
                    logger.exception("Unable to create getProperties event from received data")

            # If devicename given
            elif devicename in self.data:
                device = self.data[devicename]
                if not device.enable:
                    return
                name = root.get("name")
                if name is None:
                    try:
                        async with asyncio.TaskGroup() as tg:
                            # to all vectors of the device
                            for pname, pvector in device.items():
                                if not pvector.enable:
                                    continue
                                if self.auto_send_def:
                                    tg.create_task( pvector.send_defVector() )
                                else:
                                    # create event
                                    e = events.getProperties(devicename, pname, pvector, root)
                                    tg.create_task( self.rxevent(e) )
                    except Exception:
                        logger.exception("Unable to create getProperties event from received data")
                else:
                    pvector = device.get(name)
                    if pvector is None:
                        # name not recognised in this device
                        return
                    if not pvector.enable:
                        return
                    try:
                        if self.auto_send_def:
                            await pvector.send_defVector()
                        else:
                            # create event
                            e = events.getProperties(devicename, name, pvector, root)
                            await self.rxevent(e)
                    except Exception:
                        logger.exception("Unable to create getProperties event from received data")
            # else device not recognised, so ignore

        elif root.tag in client_tags:
            devicename = root.get("device")
            if devicename is None:
                # device not given, invalid, ignore this
                return
            elif devicename in self.data:
                if not self.data[devicename].enable:
                    return
                name = root.get("name")
                if name is None:
                    # vector name not given, invalid, ignore this
                    return
                else:
                    pvector = self.data[devicename].get(name)
                    if pvector is None:
                        # name not recognised in this device
                        return
                    e = pvector.create_event(root)
                    if e is None:
                        # no event created
                        return
                    await self.rxevent(e)


        elif root.tag in snoop_tags:
            # xml received from other devices
            # either devicename is unknown, or this data is to/from another driver.
            # So check if this driver is snooping on this device/vector
            # only forward def's and set's, not 'new' vectors which
            # do not come from a device, but only from a client to the target device.
            devicename = root.get("device")
            name = root.get("name")
            if self.snoopall:
                await self._snoophandler(root)
            elif devicename and (devicename in self.snoopdevices):
                await self._snoophandler(root)
            elif devicename and name and ((devicename, name) in self.snoopvectors):
                await self._snoophandler(root)


    async def _snoophandler(self, root):
        """Creates snoop events using data from self.readdata"""
        if self._stop:
            return
        devicename = root.get("device")
        if devicename is not None:
            # if a device name is given, check
            # it is not in this drivers devices
            if devicename in self.data:
                logger.error("Cannot snoop on a device already controlled by this driver")
                return
        try:
            if root.tag == "message":
                # create event
                event = events.message(root)
                await self._call_snoopevent(event)
            elif root.tag == "delProperty":
                # create event
                event = events.delProperty(root)
                await self._call_snoopevent(event)
            elif root.tag == "defSwitchVector":
                # create event
                event = events.defSwitchVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "setSwitchVector":
                # create event
                event = events.setSwitchVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "defLightVector":
                # create event
                event = events.defLightVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "setLightVector":
                # create event
                event = events.setLightVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "defTextVector":
                # create event
                event = events.defTextVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "setTextVector":
                # create event
                event = events.setTextVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "defNumberVector":
                # create event
                event = events.defNumberVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "setNumberVector":
                # create event
                event = events.setNumberVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "defBLOBVector":
                # create event
                event = events.defBLOBVector(root)
                await self._call_snoopevent(event)
            elif root.tag == "setBLOBVector":
                # create event
                event = events.setBLOBVector(root)
                await self._call_snoopevent(event)
        except events.EventException:
            # if an EventException is raised, it is because received data is malformed
            # so log it
            logger.exception("An exception occurred creating a snoop event")



class Device(collections.UserDict):

    """An instance of this should be created for each device controlled by this driver.
       The properties argument is a list of vectors controlling this device.
       devicedata will be an attribute dictionary of any hardware data that may be usefull.

       This object will be a mapping of vector name to vector object
    """

    def __init__(self, devicename, properties, **devicedata):
        super().__init__()

        if not devicename.isascii():
            raise ValueError("Device name should be all ascii characters.")

        # This device name
        self.devicename = devicename

        # if self.enable is False, this device ignores incoming traffic
        # and (apart from delProperty) does not transmit anything
        # from his device
        self.enable = True

        # the driver places data in this que to send data to this device
        # self.dataque = asyncio.Queue(4)

        # Every property of this device has a dataque, which is set into this dictionary
        # self.propertyquedict = {p.name:p.dataque for p in properties}

        # dictionary of optional data
        self.devicedata = devicedata

        # this will be set when the driver is created
        self.driver = None

        # self.data is a dictionary of name to vector this device owns
        for p in properties:
            p.devicename = self.devicename
            self.data[p.name] = p

        # shutdown routine sets this to True to stop coroutines
        self._stop = False


    def properties(self):
        "Returns a list of vector objects"
        return list(self.data.values())


    def shutdown(self):
        """Sets the flag self._stop to True"""
        self._stop = True


    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop


    async def send_device_message(self, message="", timestamp=None):
        """Send a message associated with this device, which the client could display.
           The timestamp should be either None or a datetime.datetime object. If the
           timestamp is None a UTC value will be inserted."""
        if not self.enable:
            # messages only sent if self.enable is True
            return
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_device_message must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('message')
        xmldata.set("device", self.devicename)
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.driver.send(xmldata)


    async def send_delProperty(self, message="", timestamp=None):
        """Sending delProperty with this device method, (as opposed to the vector send_delProperty method)
           informs the client this device is not available, it also sets a device.enable attribute to
           False, which stops any data being transmitted between the client and this device.
           Setting device.enable to True re-enables communications.
           The message argument is any appropriate string which the client could display to the user.
           The timestamp should be either None or a datetime.datetime object. If the timestamp is None
           a UTC value will be inserted."""
        tstring = timestamp_string(timestamp)
        if not tstring:
            logger.error("The timestamp given in send_delProperty must be a datetime.datetime UTC object")
            return
        xmldata = ET.Element('delProperty')
        xmldata.set("device", self.devicename)
        xmldata.set("timestamp", tstring)
        if message:
            xmldata.set("message", message)
        await self.driver.send(xmldata)
        self.enable = False


    async def devhardware(self, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'hardware' method if it is used."""
        pass


    async def devrxevent(self, event, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'rxevent' method if it is used."""
        pass


    async def devsnoopevent(self, event, *args, **kwargs):
        """As default, does nothing and is not called.

           If required, override this to handle any necessary device actions.
           This should be called from the driver 'snoopevent' method if it is used."""
        pass


    def __setitem__(self, vectorname):
        raise KeyError



class _STDINOUT():
    """This class is used to implement communications via stdin and stdout"""

    def __init__(self, driver):
        self.driver = driver
        self.connected = True
        self._remainder = b""    # Used to store intermediate data
        self._stop = False       # Gets set to True to stop communications


    @property
    def stop(self):
        "returns self._stop, being the instruction to stop the driver"
        return self._stop

    def shutdown(self):
        self._stop = True

    async def run_rx(self):
        """Called from indipydriver to get received data
           this runs continuously, checking received data"""
        # Set stdin to non-blocking mode
        flags = fcntl.fcntl(sys.stdin.fileno(), fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK)

        logger.info("Communicating via STDIN/STDOUT")

        # this runs until self._stop becomes True
        # passes incoming data to the driver
        try:
            # get block of xml.etree.ElementTree data
            # from self._xmlinput and send it to the driver
            devicenames = self.driver.keys()
            while not self._stop:
                rxdata = await self._xmlinput()
                if rxdata is None:
                    return
                # check for incoming duplicates
                if rxdata.tag in DEFTAGS:
                    devicename = rxdata.get("device")
                    if devicename is None:
                        # invalid definition
                        continue
                    if devicename in devicenames:
                        # duplicate address
                        logger.error(f"Duplicate address: Received a definition of device {devicename}")
                        self.driver.shutdown()
                        self.shutdown()
                        break
                await self.driver._readdata(rxdata)
        except Exception:
            logger.exception("Exception report from _STDINOUT.run_rx")
            raise


    async def run_tx(self, txdata):
        """Gets data to be transmitted, and transmits it out on stdout"""
        if self._stop:
            return

        if (txdata.tag == "setBLOBVector") and len(txdata):
            # txdata is a setBLOBVector containing blobs
            # send initial setBLOBVector
            startdata = _makestart(txdata)
            sys.stdout.buffer.write(startdata.encode())
            sys.stdout.buffer.flush()
            for oneblob in txdata.iter('oneBLOB'):
                bytescontent = oneblob.text.encode()
                # send start of oneblob
                startoneblob = _makestart(oneblob)
                sys.stdout.buffer.write(startoneblob.encode())
                sys.stdout.buffer.flush()
                # send content in chunks
                chunksize = 1000
                for b in range(0, len(bytescontent), chunksize):
                    byteschunk = bytescontent[b:b+chunksize]
                    sys.stdout.buffer.write(byteschunk)
                    sys.stdout.buffer.flush()
                    await asyncio.sleep(0)
                sys.stdout.buffer.write(b"</oneBLOB>")
                sys.stdout.buffer.flush()
            # send enddata
            sys.stdout.buffer.write(b"</setBLOBVector>\n")
            sys.stdout.buffer.flush()
        else:
            # its straight xml, send it out on stdout
            binarydata = ET.tostring(txdata)
            binarydata += b"\n"
            sys.stdout.buffer.write(binarydata)
            sys.stdout.buffer.flush()



    async def _xmlinput(self):
        """get data from  _datainput, parse it, and return it as xml.etree.ElementTree object
           Returns None if stop flags arises"""
        message = b''
        messagetagnumber = None
        while not self._stop:
            await asyncio.sleep(0)
            data = await self._datainput()
            # data is either None, or binary data ending in b">"
            if data is None:
                return
            if self._stop:
                return
            if not message:
                # data is expected to start with <tag, first strip any newlines
                data = data.strip()
                for index, st in enumerate(_STARTTAGS):
                    if data.startswith(st):
                        messagetagnumber = index
                        break
                    elif st in data:
                        # remove any data prior to a starttag
                        positionofst = data.index(st)
                        data = data[positionofst:]
                        messagetagnumber = index
                        break
                else:
                    # data does not start with a recognised tag, so ignore it
                    # and continue waiting for a valid message start
                    continue
                # set this data into the received message
                message = data
                # either further children of this tag are coming, or maybe its a single tag ending in "/>"
                if message.endswith(b'/>'):
                    # the message is complete, handle message here
                    try:
                        root = ET.fromstring(message.decode("us-ascii"))
                    except Exception:
                        # failed to parse the message, continue at beginning
                        message = b''
                        messagetagnumber = None
                        continue
                    # xml datablock done, return it
                    return root
                # and read either the next message, or the children of this tag
                continue
            # To reach this point, the message is in progress, with a messagetagnumber set
            # keep adding the received data to message, until an endtag is reached
            message += data
            if message.endswith(_ENDTAGS[messagetagnumber]):
                # the message is complete, handle message here
                try:
                    root = ET.fromstring(message.decode("us-ascii"))
                except Exception:
                    # failed to parse the message, continue at beginning
                    message = b''
                    messagetagnumber = None
                    continue
                # xml datablock done, return it
                return root
            # so message is in progress, with a messagetagnumber set
            # but no valid endtag received yet, so continue the loop


    async def _datainput(self):
        """Waits for binary string of data ending in > from stdin
           Returns None if stop flags arises"""
        remainder = self._remainder
        if b">" in remainder:
            # This returns with binary data ending in > as long
            # as there are > characters in self._remainder
            binarydata, self._remainder = remainder.split(b'>', maxsplit=1)
            binarydata += b">"
            return binarydata
        # As soon as there are no > characters left in self._remainder
        # get more data from stdin
        while not self._stop:
            await asyncio.sleep(0)
            indata = sys.stdin.buffer.read(100)
            if not indata:
                await asyncio.sleep(0.02)
                continue
            remainder += indata
            if b">" in indata:
                binarydata, self._remainder = remainder.split(b'>', maxsplit=1)
                binarydata += b">"
                return binarydata
